@@ -21,8 +21,9 @@ use futures::channel::{mpsc, oneshot};
 use futures::future::FutureExt;
 use futures::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
 use futures::sink::SinkExt;
-use futures::stream::StreamExt;
+use futures::stream::{FusedStream, StreamExt};
 use netsim_embed_core::{Ipv4Range, Packet, Plug};
+use std::collections::VecDeque;
 use std::fmt::{self, Display};
 use std::io::{Error, ErrorKind, Result, Write};
 use std::net::Ipv4Addr;
@@ -59,6 +60,7 @@ pub struct Machine<C, E> {
     tx: mpsc::UnboundedSender<C>,
     rx: mpsc::UnboundedReceiver<E>,
     join: Option<thread::JoinHandle<Result<()>>>,
+    buffer: VecDeque<E>,
 }
 
 impl<C, E> Machine<C, E>
@@ -83,6 +85,7 @@ where
             tx: cmd_tx,
             rx: event_rx,
             join: Some(join),
+            buffer: VecDeque::new(),
         }
     }
 }
@@ -115,7 +118,69 @@ impl<C, E> Machine<C, E> {
     }
 
     pub async fn recv(&mut self) -> Option<E> {
-        self.rx.next().await
+        if let Some(ev) = self.buffer.pop_front() {
+            Some(ev)
+        } else {
+            self.rx.next().await
+        }
+    }
+
+    pub async fn select<F, T>(&mut self, mut f: F) -> Option<T>
+    where
+        F: FnMut(&E) -> Option<T>,
+    {
+        if let Some((idx, res)) = self
+            .buffer
+            .iter()
+            .enumerate()
+            .find_map(|(idx, ev)| f(ev).map(|x| (idx, x)))
+        {
+            self.buffer.remove(idx);
+            return Some(res);
+        }
+        loop {
+            match self.rx.next().await {
+                Some(ev) => {
+                    if let Some(res) = f(&ev) {
+                        return Some(res);
+                    } else {
+                        self.buffer.push_back(ev);
+                    }
+                }
+                None => return None,
+            }
+        }
+    }
+
+    pub async fn select_draining<F, T>(&mut self, mut f: F) -> Option<T>
+    where
+        F: FnMut(&E) -> Option<T>,
+    {
+        while let Some(ev) = self.buffer.pop_front() {
+            if let Some(res) = f(&ev) {
+                return Some(res);
+            }
+        }
+        loop {
+            match self.rx.next().await {
+                Some(ev) => {
+                    if let Some(res) = f(&ev) {
+                        return Some(res);
+                    }
+                }
+                None => return None,
+            }
+        }
+    }
+
+    pub fn drain(&mut self) -> Vec<E> {
+        let mut res = self.buffer.drain(..).collect::<Vec<_>>();
+        if !self.rx.is_terminated() {
+            while let Ok(Some(x)) = self.rx.try_next() {
+                res.push(x);
+            }
+        }
+        res
     }
 
     pub fn up(&self) {
@@ -178,6 +243,7 @@ where
                         }
                     }
                 }
+                log::info!("{} (ctrl): closed", id);
                 Result::Ok(())
             }
             .fuse();
@@ -194,7 +260,7 @@ where
                     if buf[0] >> 4 != 4 {
                         continue;
                     }
-                    log::debug!("machine {}: sending packet", id.0);
+                    log::debug!("{} (reader): sending packet", id);
                     let mut bytes = buf[..n].to_vec();
                     if let Some(mut packet) = Packet::new(&mut bytes) {
                         packet.set_checksum();
@@ -203,6 +269,7 @@ where
                         break;
                     }
                 }
+                log::info!("{} (reader): closed", id);
                 Result::Ok(())
             }
             .fuse();
@@ -210,7 +277,7 @@ where
 
             let writer_task = async {
                 while let Some(packet) = rx.next().await {
-                    log::debug!("machine {}: received packet", id.0);
+                    log::debug!("{} (writer): received packet", id);
                     // can error if the interface is down
                     if let Ok(n) = iface.write_with(|iface| iface.send(&packet)).await {
                         if n == 0 {
@@ -218,14 +285,18 @@ where
                         }
                     }
                 }
+                log::info!("{} (writer): closed", id);
                 Result::Ok(())
             }
             .fuse();
             futures::pin_mut!(writer_task);
 
-            bin.stdin(Stdio::piped()).stdout(Stdio::piped());
+            bin.stdin(Stdio::piped())
+                .stdout(Stdio::piped())
+                .stderr(Stdio::piped());
             let mut child = bin.spawn()?;
             let mut stdout = BufReader::new(child.stdout.take().unwrap()).lines().fuse();
+            let mut stderr = BufReader::new(child.stderr.take().unwrap()).lines().fuse();
             let mut stdin = child.stdin.take().unwrap();
 
             let command_task = async {
@@ -236,6 +307,7 @@ where
                     writeln!(buf, "{}", cmd)?;
                     stdin.write_all(&buf).await?;
                 }
+                log::info!("{} (command): closed", id);
                 Result::Ok(())
             }
             .fuse();
@@ -256,17 +328,30 @@ where
                         println!("{} (stdout): {}", id, ev);
                     }
                 }
+                log::info!("{} (stdout): closed", id);
                 Result::Ok(())
             }
             .fuse();
             futures::pin_mut!(event_task);
 
+            let stderr_task = async {
+                while let Some(ev) = stderr.next().await {
+                    let ev = ev?;
+                    println!("{} (stderr): {}", id, ev);
+                }
+                log::info!("{} (stderr): closed", id);
+                Result::Ok(())
+            }
+            .fuse();
+            futures::pin_mut!(stderr_task);
+
             futures::select! {
                 res = ctrl_task => res?,
                 res = reader_task => res?,
                 res = writer_task => res?,
                 res = command_task => res?,
                 res = event_task => res?,
+                res = stderr_task => res?,
             }
             child.kill()
         });
